from requests import Session
from urllib.parse import urljoin

class Authenticate:
    def __init__(self, asset, asset_schema, http_proxy):
        self.asset = asset
        self.asset_meta = asset_schema.get('meta', {})
        self.params = {}

        self.session = WrappedSession()
        self.session.proxies = {
            'http_proxy': http_proxy,
            'https_proxy': http_proxy
        }
        self.session.verify = self.asset.get('verify', True)
        self.session.headers.update(self.asset.get('headers', {}))
        self.url = asset.get('url', self.asset_meta.get('url'))
        if not self.url:
            raise ValueError("No URL was provided. A URL must be provided in the asset parameters or under the meta"
                             "key in the asset schema.")

        self.authenticate(asset_schema.get('name', None))

    def authenticate(self, auth_type):
        # The asset name is the name of the function which was set by the get_asset_name function.
        if auth_type is not None:
            auth = getattr(self, auth_type)
            auth()

    def http_basic(self):
        self.session.auth = (self.asset['username'], self.asset['password'])

    def http_bearer(self):
        self.session.headers.update({
            'Authorization': f'Bearer {self.asset["token"]}'
        })

    def apikey(self):
        def handle_item(item: dict):
            if item['name'] not in self.asset:
                return
            if item['in'] == 'header':
                self.session.headers.update({
                    item['name']: self.asset[item['name']]
                })
            elif item['in'] == 'cookie':
                self.session.cookies.update({
                    item['name']: self.asset[item['name']]
                })
            elif item['in'] == 'query':
                self.params.update({
                    item['name']: self.asset[item['name']]
                })
        if isinstance(self.asset_meta['security'], list):
            for item in self.asset_meta['security']:
                handle_item(item)
        else:
            handle_item(self.asset_meta['security'])

    def oauth2_client_credentials(self):
        scopes = self.asset.get('scopes')
        if scopes:
            scopes = scopes.join(' ')
        data = {
            'client_id': self.asset['client_id'],
            'client_secret': self.asset['client_secret'],
            'scope': scopes,
            'grant_type': 'client_credentials'
        }

        token_url = self.asset.get("token_url") or urljoin(self.url, self.asset_meta['token_endpoint'])
        access_token = self.session.request("POST",
                                            token_url,
                                            data=data).json()['access_token']

        self.session.headers.update({"Authorization": "Bearer {}".format(access_token)})


class WrappedSession(Session):
    """A wrapper for requests.Session to override 'verify' property, ignoring REQUESTS_CA_BUNDLE environment variable.
    This is a workaround for https://github.com/kennethreitz/requests/issues/3829 (will be fixed in requests 3.0.0)
    Code sourced from user intgr https://github.com/kennethreitz/requests/issues/3829
    """

    def merge_environment_settings(self, url, proxies, stream, verify, *args, **kwargs):
        if self.verify is False:
            verify = False

        return super(WrappedSession, self).merge_environment_settings(url, proxies, stream, verify, *args, **kwargs)



import io
import re


FILENAME_HEADER_KEY = 'Content-Disposition'
FILENAME_REGEX = """filename=['"]?([^"']*)['"]"""


def get_filename(response):
    """ Posts a file obtained from the response.
    """
    filename = 'unknown'
    if FILENAME_HEADER_KEY in response.headers:
        match = re.search(FILENAME_REGEX, response.headers[FILENAME_HEADER_KEY])
        if match:
            filename = match.group(1)
    return filename


def get_file_object(response):
    return {
        'filename': get_filename(response),
        'file_data': io.BytesIO(response.content)
    }


def parse_attachment_output(response, output, manifest):
    if 'file' not in manifest['output']:
        return
    if manifest['output']['properties']['file']['type'] == 'array':
        output['file'] = [get_file_object(response)]
    else:
        output['file'] = get_file_object(response)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['connector_definition_runner']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'chevron>=0.14.0,<1.0.0', 'requests==2.28.0']

setup_kwargs = {
    'name': 'connector-def-runner',
    'version': '2.0.0',
    'description': 'Run connectors by schema definitions',
    'long_description': '# TODO\n',
    'author': 'Swimlane',
    'author_email': 'integrations@swimlane.com',
    'maintainer': 'Swimlane',
    'maintainer_email': 'None',
    'url': 'https://github.com/swimlane/connector-definition-runner',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import chevron
from requests import JSONDecodeError
from urllib.parse import urljoin
from .authenticate import Authenticate
from .attachments import parse_attachment_output


class Runner(Authenticate):

    def run(self, inputs=None, input_schema=None):
        self._update_headers(inputs)
        response = self.session.request(
            method=self.get_method(input_schema['meta']),
            url=urljoin(self.url, self.get_endpoint(inputs, input_schema['meta'])),
            **self.get_kwargs(inputs)
        )
        resp = self.parse_response(response, input_schema)
        return resp

    def _update_headers(self, inputs=None):
        self.session.headers.update(
            inputs.get('headers', {})
        )

    @staticmethod
    def get_method(self, input_meta=None):
        return input_meta['method']

    @staticmethod
    def get_endpoint(self, inputs=None, input_meta=None):
        # Mustache if available if not returns string as is.
        return chevron.render(input_meta['endpoint'], inputs.get('path_parameters', {}))

    def get_kwargs(self, inputs):
        # todo can we handle files automatically.
        self.params.update(inputs.get('parameters', {}))
        return {
            'params': self.params,
            'data': inputs.get('data_body'),
            'json': inputs.get('json_body'),
            'files': inputs.get('files')
        }

    def parse_response(self, response, input_schema):
        output = {
            'status_code': response.status_code,
            'response_headers': dict(response.headers),
            'reason': response.reason
        }
        file = parse_attachment_output(response, input_schema)
        if file:
            output['file'] = file

        try:
            output['json_body'] = response.json()
        except JSONDecodeError:
            if 'file' not in output:
                # Avoids the duplication of the file content in response_text
                output['response_text'] = response.text
        return output
